/*
 * me9000.c --  Meilhaus Electronic ME-9000 serial driver.
 *
 * Copyright (C) 2017 Meilhaus Electronic GmbH.
 *
 * Based on Linux kernel's drivers/serial/8250.c , /8250_pci.c and modifications by Exar Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 *  for		: Linux kernel 4.x and newer
 *  date	: May 2017
 *  version	: 1.0.5
 *  author	: Jan Henrik Sawatzki <support@meilhaus.com>
 *
 */
 
 /*
 *  date	: March 2018
 *  version	: 1.0.7
 *  author	: Gert Fabritius <gert.fabritius@frogsure.de>
 *  Improve functionality
 */
 
 /*
 *  date	: March 2019
 *  version	: 1.0.8
 *  author	: Gert Fabritius <gert.fabritius@frogsure.de>
 *  Adapt to new Timer API -- kernel > 4,14,0 (e.g. Ubuntu 16.04.6 LTS)
 */

// Set this parameter to 1 to enabled internal loopback
#define ENABLE_INTERNAL_LOOPBACK						0

/*
 * Set this parameter to 1 to enable Debug mode
 * The Driver enables the internal loopback under debug mode
 * To disable internal loopback go to me9000_set_termios
 */
#define DEBUG 		0
#define DEBUG_F 	0

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/pci.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include <linux/tty_flip.h>
#include <linux/serial_reg.h>
#include <linux/serial.h>
#include <linux/serial_core.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/byteorder.h>
#include <asm/serial.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "linux/version.h"

#define _INLINE_ inline



#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 8, 0)
#define __devinitdata
#define __devinit
#define __devexit
#define __devexit_p
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 8, 0)
struct serial_uart_config {
	char *name;
	int dfl_xmit_fifo_size;
	int flags;
};
#endif

#define FIFO_ON 		1

/*
* Overwrite or define values from serial_reg.h (19.3.2019)
*/
#define UART_EXAR_DVID		0x8d	/* Device identification */
#define UART_EXAR_FCTR		0x08	/* Feature Control Register */
#define UART_FCTR_EXAR_IRDA	0x10	/* IrDa data encode select */
#define UART_FCTR_EXAR_485	0x20	/* Auto 485 half duplex dir ctl */
#define UART_FCTR_EXAR_TRGA	0x00	/* FIFO trigger table A */
#define UART_FCTR_EXAR_TRGB	0x40	/* FIFO trigger table B */
#define UART_FCTR_EXAR_TRGC	0x80	/* FIFO trigger table C */
#define UART_FCTR_EXAR_TRGD	0xc0	/* FIFO trigger table D programmable */

/*
 * Definitions for PCI support.
 */
#define FL_BASE_MASK									0x0007
#define FL_BASE0										0x0000
#define FL_GET_BASE(x)									(x & FL_BASE_MASK)

#define NR_PORTS										256

#define ME9000_MAJOR									30
#define ME9000_MINOR									0

#define PCI_VENDOR_ID_MEILHAUS							0x1402

#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x0_S		0x9000
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x1_S		0x9001
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x0_S		0x9010
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x1_S		0x9011
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x2_S		0x9012
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x0_S		0x9020
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x1_S		0x9021
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x2_S		0x9022
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x3_S		0x9023
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x0_S		0x9030
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x1_S		0x9031
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x2_S		0x9032
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x3_S		0x9033
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x4_S		0x9034
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x0_S		0x9040
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x1_S		0x9041
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x2_S		0x9042
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x3_S		0x9043
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x4_S		0x9044
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x5_S		0x9045
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x0_S		0x9050
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x1_S		0x9051
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x2_S		0x9052
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x3_S		0x9053
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x4_S		0x9054
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x5_S		0x9055
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x6_S		0x9056
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x0_S		0x9060
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x1_S		0x9061
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x2_S		0x9062
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x3_S		0x9063
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x4_S		0x9064
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x5_S		0x9065
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x6_S		0x9066
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x7_S		0x9067
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x8_485x0_S		0x9070
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x1_S		0x9071
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x2_S		0x9072
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x3_S		0x9073
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x4_S		0x9074
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x5_S		0x9075
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x6_S		0x9076
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x7_S		0x9077
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x8_S		0x9078
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x0_F		0x9080
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x1_F		0x9081
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x0_F		0x9090
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x1_F		0x9091
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x2_F		0x9092
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x0_F		0x90A0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x1_F		0x90A1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x2_F		0x90A2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x3_F		0x90A3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x0_F		0x90B0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x1_F		0x90B1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x2_F		0x90B2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x3_F		0x90B3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x4_F		0x90B4
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x0_F		0x90C0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x1_F		0x90C1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x2_F		0x90C2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x3_F		0x90C3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x4_F		0x90C4
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x5_F		0x90C5
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x0_F		0x90D0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x1_F		0x90D1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x2_F		0x90D2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x3_F		0x90D3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x4_F		0x90D4
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x5_F		0x90D5
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x6_F		0x90D6
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x0_F		0x90E0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x1_F		0x90E1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x2_F		0x90E2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x3_F		0x90E3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x4_F		0x90E4
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x5_F		0x90E5
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x6_F		0x90E6
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x7_F		0x90E7
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x8_485x0_F		0x90F0
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x1_F		0x90F1
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x2_F		0x90F2
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x3_F		0x90F3
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x4_F		0x90F4
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x5_F		0x90F5
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x6_F		0x90F6
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x7_F		0x90F7
#define PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x8_F		0x90F8

/*
 * The special register set for XR17V158 UARTs.
 */
//#define XR17V158_UART_XOFF1								0xC0
//#define XR17V158_UART_XON1								0xE0
#define XR17V158_UART_XOFF1								0x0C
#define XR17V158_UART_XON1								0x0E
#define XR17V158_UART_XOFF2								0x0D
#define XR17V158_UART_XON2								0x0F
#define XR17V158_EXTENDED_TXTRG							0x0A
#define XR17V158_EXTENDED_RXTRG							0x0B
#define XR17V158_TXFIFO_CNT								0x0A
#define XR17V158_RXFIFO_CNT								0x0B
#define XR17V158_INT									0x80
#define XR17V158_8XMODE									0x88
#define XR17V158_4XMODE									0x89
#define XR17V158_SLEEP									0x8B
#define XR17V158_DREV									0x8C
#define XR17V158_DVID									0x8D

#define XR17V158_FCT_TX_FIFO_TRIGGER					0x30
#define XR17V158_FCT_RX_FIFO_TRIGGER					0xC0

#define XR17V158_UART_MCR_DTRorRTS						0x04
#define XR17V158_FCTR_RS485								0x20
#define XR17V158_FCTR_RTS_DELAY_NO						0x00
#define XR17V158_FCTR_RTS_DELAY_4						0x01
#define XR17V158_FCTR_RTS_DELAY_6						0x02
#define XR17V158_FCTR_RTS_DELAY_8						0x03
#define XR17V158_FCTR_RTS_DELAY_8_						0x04
#define XR17V158_FCTR_RTS_DELAY_12						0x0C
#define XR17V158_FCTR_RTS_DELAY_16						0x05
#define XR17V158_FCTR_RTS_DELAY_20						0x0D
#define XR17V158_FCTR_RTS_DELAY_24						0x06
#define XR17V158_FCTR_RTS_DELAY_28						0x0E
#define XR17V158_FCTR_RTS_DELAY_32						0x07
#define XR17V158_FCTR_RTS_DELAY_36						0x0F
#define XR17V158_FCTR_RTS_DELAY_40						0x08
#define XR17V158_FCTR_RTS_DELAY_44						0x09
#define XR17V158_FCTR_RTS_DELAY_48						0x0A
#define XR17V158_FCTR_RTS_DELAY_52						0x0B

#define XR17V158_EFR_TRANSMIT_XON1_XOFF1				0x08
#define XR17V158_EFR_COMPARE_XON1_XOFF1					0x02

//This delay is used
#define XR17V158_FCTR_RTS_DELAY							XR17V158_FCTR_RTS_DELAY_8

#define XR17V158_MPIOINT_7_0							0x8F
#define XR17V158_MPIOLVL_7_0							0x90
#define XR17V158_MPIO3T_7_0								0x91
#define XR17V158_MPIOINV_7_0							0x92
#define XR17V158_MPIOSEL_7_0							0x93

#define XR17V158_RX_OFFSET								0x100
#define XR17V158_TX_OFFSET 								0x100

/*
 * ME9000 ioctls
 */
#define TIOCGMPIO										0x5490
#define TIOCSMPIO										0x5491
#define EXAR_READ_REG									0x5492
#define EXAR_WRITE_REG									0x5493
#define EXAR_SET_NON_STANDARD_BAUDRATE					0x5494

#define MPIO0											0x01
#define MPIO1											0x02
#define MPIO2											0x04
#define MPIO3											0x08
#define MPIO4											0x10
#define MPIO5											0x20
#define MPIO6											0x40
#define MPIO7											0x80

#define MPIOVAL											0x100
#define MPIOINV											0x200
#define MPIODIR											0x400

#define DIVISOR_CHANGED									0

#define PCI_NUM_BAR_RESOURCES							6

#define get_from_user(arg1, arg2) get_user(arg1, (unsigned int *)arg2)

struct serial_private {
	struct pci_dev *dev;
	unsigned int nr;
	void __iomem *remapped_bar[PCI_NUM_BAR_RESOURCES];
	struct pci_serial_quirk *quirk;
	int uart_index[NR_PORTS];
	int line[0];
};

struct pciserial_board {
	unsigned int flags;
	unsigned int num_ports;
	unsigned int base_baud;
	unsigned int uart_offset;
	unsigned int reg_shift;
	unsigned int first_offset;
};

/*
 * init function returns:
 * > 0 - number of ports
 * = 0 - use board->num_ports
 * < 0 - error
 */
struct pci_serial_quirk {
	u32 vendor;
	u32 device;
	u32 subvendor;
	u32 subdevice;
	int (*init)(struct pci_dev *dev);
	int (*setup)(struct serial_private *, const struct pciserial_board *, struct uart_port *, int);
	void (*exit)(struct pci_dev *dev);
};

/*
 * This is the configuration table for all of the PCI serial boards
 * which we support. It is directly indexed by the mepci_board_num_t enum
 * value, which is encoded in the pci_device_id PCI probe table's
 * driver_data member.
 *
 * The makeup of these names are:
 * pbn_bn{_bt}_n_baud
 *
 * bn = PCI BAR number
 * bt = Index using PCI BARs
 * n = number of serial ports
 * baud = baud rate
 */
enum mepci_board_num_t {
	pbn_me9000 = 0,
};

static struct pciserial_board mepciserial_boards[] __devinitdata = {
	/*
	 * Meilhaus ME-9000
	 */
	[pbn_me9000] = {
		.flags = FL_BASE0,
		.base_baud = 921600, //uart clock = 921600 * 16 = 14745600
		.uart_offset = 0x200,
		.reg_shift = 0,
		.first_offset = 0,
	},
};

/*
 * Configuration:
 * share_irqs - whether we pass SA_SHIRQ to request_irq(). This option is unsafe when used on edge-triggered interrupts.
 */
#define SERIALEXAR_SHARE_IRQS	1
unsigned int share_irqs = SERIALEXAR_SHARE_IRQS;

/*
 * Debugging.
 */
#if DEBUG
#define DEBUG_AUTOCONF(fmt...) printk(fmt)
#else
#define DEBUG_AUTOCONF(fmt...) do { } while (0)
#endif

#if DEBUG
#define DEBUG_INTR(fmt...) printk(fmt)
#else
#define DEBUG_INTR(fmt...) do { } while (0)
#endif

#if DEBUG_F
#define DEBUG_FUNC(fmt...) printk(fmt)
#else
#define DEBUG_FUNC(fmt...) do { } while (0)
#endif

#if DEBUG_F
#define DEBUG_FUNC_S(fmt...) printk("%s start\n", fmt)
#else
#define DEBUG_FUNC_S(fmt...) do { } while (0)
#endif

#if DEBUG_F
#define DEBUG_FUNC_E(fmt...) printk("%s end\n", fmt)
#else
#define DEBUG_FUNC_E(fmt...) do { } while (0)
#endif

#define PASS_LIMIT	256

/*
 * We default to IRQ0 for the "no irq" hack. Some
 * machine types want others as well - they're free
 * to redefine this in their header file.
 */
#define is_real_interrupt(irq)	((irq) != 0)
#define MSR_SAVE_FLAGS UART_MSR_ANY_DELTA

#define LSR_SAVE_FLAGS UART_LSR_BRK_ERROR_BITS
struct uart_me9000_port {
	struct uart_port port;
	struct timer_list timer; /* "no irq" timer */
	struct list_head list; /* ports on this IRQ */
	unsigned short capabilities; /* port capabilities */
	unsigned short bugs; /* port bugs */
	unsigned int tx_loadsz; /* transmit fifo load size */
	unsigned char acr;
	unsigned char ier;
	unsigned char lcr;
	unsigned char mcr;
	unsigned char mcr_mask; /* mask of user bits */
	unsigned char mcr_force; /* mask of forced bits */

	unsigned char lsr_saved_flags;
	unsigned char msr_saved_flags;

	unsigned short deviceid;
	unsigned char channelnum;

	/*
	 * We provide a per-port pm hook.
	 */
	void (*pm)(struct uart_port *port, unsigned int state, unsigned int old);
};

struct irq_info {
	struct hlist_node node;
	int irq;
	spinlock_t lock; /* Protects list not the hash */
	struct list_head *head;
};

#define NR_IRQ_HASH		32 /* Can be adjusted later */
static struct hlist_head irq_lists[NR_IRQ_HASH];
static DEFINE_MUTEX(hash_mutex); /* Used to walk the hash */

/*
 * Here we define the default xmit fifo size used for each type of UART.
 */
#define PORT_MAX_ME	2
#define XR17V158	1 // the third entry that is [2] in the array

static const struct serial_uart_config uart_config[PORT_MAX_ME + 1] = {
	{ "Unknown", 1, 0 },
	{ "XR17V158", 64, 0 },
};

#define BOTH_EMPTY	(UART_LSR_TEMT | UART_LSR_THRE)

static int quot_coeff = 16;

struct meioctl_rw_reg {
	unsigned char reg;
	unsigned char regvalue;
};

static DEFINE_MUTEX(serial_mutex);

static struct uart_me9000_port me9000_ports[NR_PORTS];

#define me9000_CONSOLE	NULL

static struct uart_driver me9000_uart_driver = {
	.owner = THIS_MODULE,
	.driver_name = "me9000",
	.dev_name = "ttyME",
	.major = ME9000_MAJOR,
	.minor = ME9000_MINOR,
	.nr = NR_PORTS,
	.cons = me9000_CONSOLE,
};

static struct pci_device_id me9000_pci_tbl[] = {
	/* Meilhaus ME-9000 boards */
	{
		PCI_VENDOR_ID_MEILHAUS, PCI_ANY_ID,
		PCI_ANY_ID, PCI_ANY_ID,
		0, 0,
		pbn_me9000
	},

	{ 0, }
};

/*
 * Meilhaus Electronic GmbH
 * Meilhaus ME-9000 series has an explosion of boards, and to avoid the PCI table from
 * growing *huge*, we use this function to collapse the 88 entries
 * in the PCI table into one, for sanity's and compactness's sake.
 */
static unsigned short me9000_one_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x1_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x1_F, 0
};

static unsigned short me9000_two_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x2_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x2_F, 0
};

static unsigned short me9000_three_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x3_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x3_F, 0
};

static unsigned short me9000_four_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x3_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x4_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x3_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x4_F, 0
};

static unsigned short me9000_five_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x3_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x4_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x5_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x3_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x4_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x5_F, 0
};

static unsigned short me9000_six_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x3_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x4_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x5_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x6_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x3_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x4_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x5_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x6_F, 0
};

static unsigned short me9000_seven_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x3_S, 	PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x4_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x5_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x6_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x7_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x3_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x4_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x5_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x6_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x7_F, 0
};

static unsigned short me9000_eight_port[] = {
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x8_485x0_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x1_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x2_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x3_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x4_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x5_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x6_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x7_S, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x8_S,
	PCI_DEVICE_ID_MEILHAUS_ME9000_232x8_485x0_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x7_485x1_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x6_485x2_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x5_485x3_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x4_485x4_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x3_485x5_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x2_485x6_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x1_485x7_F, PCI_DEVICE_ID_MEILHAUS_ME9000_232x0_485x8_F, 0
};

static struct me9000_struct {
	int num;
	unsigned short *ids;
} me9000_data[] = {
	{ 1, me9000_one_port },
	{ 2, me9000_two_port },
	{ 3, me9000_three_port },
	{ 4, me9000_four_port },
	{ 5, me9000_five_port },
	{ 6, me9000_six_port },
	{ 7, me9000_seven_port },
	{ 8, me9000_eight_port },
	{ 0, 0 }
};

static _INLINE_ unsigned int serial_in(struct uart_me9000_port *up, int offset)
{
	unsigned char value;
	unsigned long membase = up->port.membase;
	if (offset >= 0x80 && offset <=0x93) {
		// device config register
		membase = membase >> 12;
		membase = membase << 12;
	}
	value= readb(membase + offset);
	DEBUG_INTR(KERN_INFO "readb 0x%x - 0x%x\n", membase + offset, value);
	return value;
}
static _INLINE_ unsigned long serial_in_l(struct uart_me9000_port *up, int offset)
{
	unsigned long value;
	unsigned long membase = up->port.membase;
	if (offset >= 0x80 && offset <=0x93) {
		// device config register
		membase = membase >> 12;
		membase = membase << 12;
	}
	value= readl(membase + offset);

	DEBUG_INTR(KERN_INFO "readl 0x%x - 0x%x\n", membase + offset, value);
	return value;
}

static _INLINE_ void serial_out(struct uart_me9000_port *up, int offset, int value)
{
	unsigned long membase = up->port.membase;
	if (offset >= 0x80 && offset <=0x93) {
		// device config register
		membase = membase >> 12;
		membase = membase << 12;
	}
	writeb(value, membase + offset);
	DEBUG_INTR(KERN_INFO "writeb 0x%x - 0x%x\n", membase + offset, value);
}

static _INLINE_ void serial_out_l(struct uart_me9000_port *up, int offset, unsigned long value)
{
	unsigned long membase = up->port.membase;
	if (offset >= 0x80 && offset <=0x93) {
		// device config register
		membase = membase >> 12;
		membase = membase << 12;
	}
	writel(value, membase + offset);
	DEBUG_INTR(KERN_INFO "writel 0x%x - 0x%x\n", membase + offset, value);
}

static int pci_me9000_rs485_config(struct uart_port *port, struct serial_rs485 *rs485)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned int fctr_reg = 0;

	if (!rs485)
		rs485 = &port->rs485;
	else if (rs485->flags & SER_RS485_ENABLED)
		memset(rs485->padding, 0, sizeof(rs485->padding));
	else
		memset(rs485, 0, sizeof(*rs485));

	rs485->flags &= SER_RS485_ENABLED;

	fctr_reg = serial_in(up, UART_EXAR_FCTR);
	DEBUG_INTR(KERN_INFO "pci_me9000_rs485_config: FCTR=0x%x", fctr_reg);

	unsigned int  mcr_reg =  serial_in(up, UART_MCR);
	if (rs485->flags & SER_RS485_ENABLED) {
		/* Enable RTS H/W control mode */
		DEBUG_INTR(KERN_INFO "Enable rs485\n");
		serial_out(up, UART_EXAR_FCTR, fctr_reg | XR17V158_FCTR_RS485);

		//TODO rs485 delay after send MSR [7:4] page 50 in datasheet
		serial_out(up, UART_MCR, mcr_reg | XR17V158_UART_MCR_DTRorRTS); //use DTR for Auto RS-485 Control
		rs485->delay_rts_after_send = 0;
		rs485->delay_rts_before_send = 0;
	} else {
		/* Disable RTS H/W control mode */
		DEBUG_INTR(KERN_INFO "Disable rs485\n");
		serial_out(up, UART_MCR, mcr_reg & ~XR17V158_UART_MCR_DTRorRTS); //unset DTR for Auto RS-485 Control
		serial_out(up, UART_EXAR_FCTR, fctr_reg & ~XR17V158_FCTR_RS485);
	}

	if (rs485 != &port->rs485)
		port->rs485 = *rs485;

	DEBUG_FUNC_E(__FUNCTION__);

	return 0;
}

static int setup_port(struct serial_private *priv, struct uart_port *port, int bar, int offset, int regshift, int idx)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct pci_dev *dev = priv->dev;
	unsigned long base, len;

	DEBUG_INTR(KERN_INFO "setup port 0x%x bar =x%x\n", priv->dev->device, bar);
	if (bar >= PCI_NUM_BAR_RESOURCES)
		return -EINVAL;

	base = pci_resource_start(dev, bar);

	if (pci_resource_flags(dev, bar) & IORESOURCE_MEM) {
		len = pci_resource_len(dev, bar);

		DEBUG_INTR(KERN_INFO "base 0x%x len =x%x\n", base, len);
		if (!priv->remapped_bar[bar])
			priv->remapped_bar[bar] = ioremap(base, len);
		if (!priv->remapped_bar[bar])
			return -ENOMEM;

		port->iotype = UPIO_MEM;
		port->iobase = 0;

		DEBUG_INTR(KERN_INFO "port->rs485_config = pci_me9000_rs485_config\n");
		port->rs485_config = pci_me9000_rs485_config;
		port->mapbase = base + offset;
		port->membase = priv->remapped_bar[bar] + offset;
		port->regshift = regshift;

		/*
		 * Setup Multipurpose Input/Output pins.
		 */
		if (idx == 0) {
			writeb(0x00, port->membase + XR17V158_MPIOINT_7_0);
			writeb(0x00, port->membase + XR17V158_MPIOLVL_7_0);
			writeb(0x00, port->membase + XR17V158_MPIO3T_7_0);
			writeb(0x00, port->membase + XR17V158_MPIOINV_7_0);
			writeb(0x00, port->membase + XR17V158_MPIOSEL_7_0);
		}
	} else {
		//Exar's got to be memory mapped. some hardware reading error?
		return -EINVAL;
	}
	DEBUG_INTR(KERN_INFO "membase 0x%x\n", port->membase);
	DEBUG_FUNC_E(__FUNCTION__);
	return 0;
}

static int pci_me9000_init(struct pci_dev *dev)
{
	DEBUG_FUNC_S(__FUNCTION__);
	unsigned short *ids;
	int i, j;


	for (i = 0; me9000_data[i].num; i++) {
		ids = me9000_data[i].ids;
			for (j = 0; ids[j]; j++)
				if(dev->device == ids[j])
					return me9000_data[i].num;
	}
	DEBUG_FUNC_E(__FUNCTION__);
	return 0;
}

static int pci_default_setup(struct serial_private *priv, const struct pciserial_board *board, struct uart_port *port, int idx)
{
	DEBUG_FUNC_S(__FUNCTION__);
	unsigned int bar, offset = board->first_offset;

	bar = FL_GET_BASE(board->flags);
	offset += idx * board->uart_offset;
	printk(KERN_INFO "Meilhaus ME-9000 device 0x%x\n", priv->dev->device);
	DEBUG_INTR(KERN_INFO "pci_default_setup end call setup_port\n");
	return setup_port(priv, port, bar, offset, board->reg_shift, idx);
}

/*
 * Master list of serial port init/setup/exit quirks.
 * This does not describe the general nature of the port.
 * (ie, baud base, number and location of ports, etc)
 *
 * This list is ordered alphabetically by vendor then device.
 * Specific entries must come before more generic entries.
 */
static struct pci_serial_quirk pci_serial_quirks[] = {
	/*
	 * Meilhaus Electronic GmbH ME-9000 series boards
	 *
	 */
	{
		.vendor = PCI_VENDOR_ID_MEILHAUS,
		.device = PCI_ANY_ID,
		.subvendor = PCI_ANY_ID,
		.subdevice = PCI_ANY_ID,
		.init = pci_me9000_init,
		.setup = pci_default_setup,
	},

};

static inline int quirk_id_matches(u32 quirk_id, u32 dev_id)
{
	return quirk_id == PCI_ANY_ID || quirk_id == dev_id;
}

static struct pci_serial_quirk *find_quirk(struct pci_dev *dev)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct pci_serial_quirk *quirk;

	for (quirk = pci_serial_quirks; ; quirk++)
		if (quirk_id_matches(quirk->vendor, dev->vendor) && quirk_id_matches(quirk->device, dev->device))
			break;
	return quirk;
}

static void me9000_stop_tx(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned int  lcr;


	if (up->ier & UART_IER_THRI) {
		up->ier &= ~UART_IER_THRI;
		lcr = serial_in(up, UART_LCR);
		serial_out(up, UART_LCR, lcr & ~UART_LCR_DLAB); // ensure LCR bit-7=0 before writing IER
		serial_out(up, UART_IER, up->ier);
		serial_out(up, UART_LCR, lcr);
	}
	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_start_tx(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned int  lcr;

	if (!(up->ier & UART_IER_THRI)) {
		up->ier |= UART_IER_THRI;
		lcr = serial_in(up, UART_LCR);
		serial_out(up, UART_LCR, lcr & ~UART_LCR_DLAB); // ensure LCR bit-7=0 before writing IER
		serial_out(up, UART_IER, up->ier);
		serial_out(up, UART_LCR, lcr);
	}
	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_stop_rx(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned int  lcr;

	up->ier &= ~UART_IER_RLSI;
	up->port.read_status_mask &= ~UART_LSR_DR;
	lcr = serial_in(up, UART_LCR);
	serial_out(up, UART_LCR, lcr & ~UART_LCR_DLAB); // ensure LCR bit-7=0 before writing IER
	serial_out(up, UART_IER, up->ier);
	serial_out(up, UART_LCR, lcr);
	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_enable_ms(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	unsigned int  lcr;
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;

	up->ier |= UART_IER_MSI;
	lcr = serial_in(up, UART_LCR);
	serial_out(up, UART_LCR, lcr & ~UART_LCR_DLAB); // ensure LCR bit-7=0 before writing IER
	serial_out(up, UART_IER, up->ier);
	serial_out(up, UART_LCR, lcr);
	DEBUG_FUNC_E(__FUNCTION__);
}

static void receive_chars(struct uart_me9000_port *up, unsigned int *status)
{
	DEBUG_FUNC_S(__FUNCTION__);
	
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 9, 0)
	struct uart_port *port = &up->port;
#else
	struct tty_struct *tty = up->port.state->port.tty;
#endif

	unsigned char ch[256], lsr = *status;
	char flag;
	int i, j, datasize_in_fifo, port_index;
	struct tty_port *tport = &up->port.state->port;

	DEBUG_INTR("Atomic 0x%x start\n", up->channelnum);

#if FIFO_ON
	datasize_in_fifo = serial_in(up, XR17V158_RXFIFO_CNT);
	while (datasize_in_fifo != serial_in(up, XR17V158_RXFIFO_CNT))
		/*Read Receive Fifo count until we get read same value twice*/
		datasize_in_fifo = serial_in(up, XR17V158_RXFIFO_CNT);
#else
	datasize_in_fifo = 1;
#endif

	port_index = up->port.line;
	flag = TTY_NORMAL;

	DEBUG_INTR("Receive Chars Status: 0x%x\n", lsr);
	if (unlikely(lsr & (UART_LSR_BI | UART_LSR_PE | UART_LSR_FE | UART_LSR_OE)))
	{
		/*
		* Mask off conditions which should be ignored.
		*/
		DEBUG_INTR("Ignore conditions\n");
		lsr &= up->port.read_status_mask;

		DEBUG_INTR("New lsr: 0x%x\n", lsr);

		if (lsr & UART_LSR_BI)
		{
			DEBUG_INTR("handling break....\n");
			flag = TTY_BREAK;
		}
		else if (lsr & UART_LSR_PE)
		{
			flag = TTY_PARITY;
			printk("handling port<%d> Parity error....\n", port_index);

		}
		else if (lsr & UART_LSR_FE)
		{
			DEBUG_INTR("handling Frame error....\n");
			flag = TTY_FRAME;
		}
	}

#if FIFO_ON
    //memcpy_fromio(ch, up->port.membase + XR17V158_RX_OFFSET, datasize_in_fifo);

    long d1;
    j=0;
    for(i = 0; i < datasize_in_fifo/4; i++)
    {
    	j = i*4;
    	d1 = serial_in_l(up, XR17V158_RX_OFFSET + j);

    	ch[j] = (d1  & 0x000000ffUL);
    	ch[j+1] = (d1  & 0x0000ff00UL) >> 8;
    	ch[j+2] = (d1  & 0x00ff0000UL) >> 16;
    	ch[j+3] = (d1  & 0xff000000UL) >> 24;
    	j = j + 4;
    }

    // get rest data (multiple 4 to end)
    for(i = j; i < datasize_in_fifo; i++)
    {
    	d1 = serial_in(up, XR17V158_RX_OFFSET + i);
    	ch[i] = (d1  & 0x000000ffUL);
    }

	//d1 = serial_in_l(up, XR17V158_RX_OFFSET + 0x80);
#else
    ch[0] = serial_in(up, UART_RX);
#endif

    DEBUG_INTR("count_rx/add: 0x%x/0x%x \n", up->port.icount.rx, datasize_in_fifo);

	up->port.icount.rx += datasize_in_fifo;

	int t2 = 1;

	int free;
	if (t2 == 1)
	{
	for(i = 0; i < datasize_in_fifo; i++)
	{
		DEBUG_INTR("i/ch: %d/%d/0x%x/%c\n", i, ch[i], ch[i], ch[i]);
		//fab
		//if (uart_handle_sysrq_char(&up->port, ch[i]))continue;
		DEBUG_INTR("insert ch: 0x%x - %d -- lsr/flag/mask: 0x%x/0x%x/0x%x\n", ch[i], ch[i], lsr, flag,up->port.ignore_status_mask);
		//fab
		//uart_insert_char(&up->port, lsr, UART_LSR_OE, ch[i], flag);
		tty_insert_flip_char(tport, ch[i], flag);

		free = tty_buffer_space_avail(tport);
		DEBUG_INTR("buffer used/limit/free: 0x%x/0x%x/0x%x\n", tport->buf.mem_used, tport->buf.mem_limit, free);
		//tty_insert_flip_char(tport, ch[i], TTY_NORMAL);
	}
	} else {
		free = tty_buffer_space_avail(tport);
		DEBUG_INTR("buffer used/limit/free: 0x%x/0x%x/0x%x\n", tport->buf.mem_used, tport->buf.mem_limit, free);

		tty_insert_flip_string_flags(tport, ch, flag, datasize_in_fifo);

		free = tty_buffer_space_avail(tport);
		DEBUG_INTR("buffer used/limit/free: 0x%x/0x%x/0x%x\n", tport->buf.mem_used, tport->buf.mem_limit, free);
	}

	DEBUG_INTR("Atomic 0x%x end\n", up->channelnum);

	//fab
	//spin_unlock(&up->port.lock);

#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 9, 0)
	DEBUG_INTR("flush flip buffer\n");
	//tty_flip_buffer_push(&port->state->port);
	tty_flip_buffer_push(tport);
#else
	tty_flip_buffer_push(tty);
#endif

	//fab
	//spin_lock(&up->port.lock);

	DEBUG_FUNC_E(__FUNCTION__);
}

static void transmit_chars(struct uart_me9000_port *up)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct circ_buf *xmit = &up->port.state->xmit;
	int count, bytes_in_fifo, tmp_count, i, j;

	if (up->port.x_char) {
		DEBUG_INTR("transmit_chars xon/xoff 0x%x\n", up->port.x_char);
		serial_out(up, UART_TX, up->port.x_char);
		up->port.icount.tx++;
		up->port.x_char = 0;
		return;
	}
	if (uart_tx_stopped(&up->port)) {
		DEBUG_INTR("transmit_chars tx stooped.\n");
		me9000_stop_tx(&up->port);
		return;
	}
	if (uart_circ_empty(xmit)) {
		DEBUG_INTR("transmit_chars circ_empty.\n");
		me9000_stop_tx(&up->port);
		return;
	}

	bytes_in_fifo = serial_in(up, XR17V158_TXFIFO_CNT);
	// read the fifo count untill we get the same value twice
	while (bytes_in_fifo != serial_in(up, XR17V158_TXFIFO_CNT))
		bytes_in_fifo = serial_in(up, XR17V158_TXFIFO_CNT);

	// how much buffer is availabe now to write?
	count = up->port.fifosize - bytes_in_fifo;

	DEBUG_INTR("transmit chars countToWrite/fifosize/bytesInFifo - 0x%x/0x%x/0x%x\n", count, up->port.fifosize, bytes_in_fifo);
	if (uart_circ_chars_pending(xmit) < count)
		count = uart_circ_chars_pending(xmit);

	do
	{
		// if the count is more than (tail to end of the buffer), transmit only the rest here.
		// tail+tmp&(UART_XMIT_SIZE-1) will reset the tail to the starting of the circular buffer
		if ( ((xmit->tail + count) & (UART_XMIT_SIZE-1)) < xmit->tail)
		{
			tmp_count = UART_XMIT_SIZE - xmit->tail;
			DEBUG_INTR("transmit_chars put (full) 0x%x to fifo.\n", tmp_count);
		}
		else
		{
			tmp_count = count;
			DEBUG_INTR("transmit_chars put (last) 0x%x to fifo.\n", tmp_count);
		}

		//memcpy_toio(up->port.membase + XR17V158_TX_OFFSET, &(xmit->buf[xmit->tail]), tmp_count);

		long d1;
		j=0;
		for(i = 0; i < tmp_count/4; i+=1)
		{
			j=i*4;
			d1 = (xmit->buf[xmit->tail + j] & 0x000000ffUL) |
				(xmit->buf[xmit->tail + j + 1]<<8   & 0x0000ff00UL) |
				(xmit->buf[xmit->tail + j + 2]<<16  & 0x00ff0000UL) |
				(xmit->buf[xmit->tail + j + 3]<<24  & 0xff000000UL);

			serial_out_l(up, XR17V158_TX_OFFSET + j, d1);
			j=j+4;
		}

		//put rest (multiple 4 to end)
		for(i = j; i < tmp_count; i+=1)
		{
			serial_out(up, XR17V158_TX_OFFSET + i, xmit->buf[xmit->tail + i]);
		}
		xmit->tail += tmp_count;
		xmit->tail &= (UART_XMIT_SIZE-1);
		up->port.icount.tx += tmp_count;
		count -= tmp_count;

	}while (count > 0);

	if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS)
		uart_write_wakeup(&up->port);

	DEBUG_INTR("THRE...\n");

	if (uart_circ_empty(xmit))
		me9000_stop_tx(&up->port);

	DEBUG_FUNC_E(__FUNCTION__);
}

static unsigned int check_modem_status(struct uart_me9000_port *up)
{
	DEBUG_FUNC_S(__FUNCTION__);

	unsigned int status = serial_in(up, UART_MSR);

	status |= up->msr_saved_flags;
	up->msr_saved_flags = 0;
	if (status & UART_MSR_ANY_DELTA && up->ier & UART_IER_MSI &&
		up->port.state != NULL) {
		if (status & UART_MSR_TERI)
			up->port.icount.rng++;
		if (status & UART_MSR_DDSR)
			up->port.icount.dsr++;
		if (status & UART_MSR_DDCD)
			uart_handle_dcd_change(&up->port, status & UART_MSR_DCD);
		if (status & UART_MSR_DCTS)
			uart_handle_cts_change(&up->port, status & UART_MSR_CTS);

		wake_up_interruptible(&up->port.state->port.delta_msr_wait);
	}

	DEBUG_INTR("check_modem_status status=0x%x.\n");
	DEBUG_FUNC_E(__FUNCTION__);
	return status;
}

/*
 * This handles the interrupt from one port.
 */
static void me9000_handle_port(struct uart_me9000_port *up)
{
	unsigned int status;
	unsigned long flags;

	DEBUG_FUNC_S(__FUNCTION__);

	spin_lock_irqsave(&up->port.lock, flags);

	status = serial_in(up, UART_LSR);

	DEBUG_INTR("status = 0x%x\n", status);

	//fab
	int chan_nr = up->channelnum & 0x7;
	long status_reg = serial_in_l(up, XR17V158_INT);
	long channel_status = status_reg >> 8; //ignore INT0
	channel_status = (channel_status >> (chan_nr * 3)) & 0x07UL; // 3 bit per channel

	long int_mask = 3 << (chan_nr * 3); // 3 bit per channel
	int_mask = int_mask << 8; //ignore INT0 byte

	DEBUG_INTR("INT/CHANNAL/CHANNEL-STATUS = 0x%x/0x%x/0x%x\n", status_reg, chan_nr, channel_status);

	//if ((status & (UART_LSR_DR | UART_LSR_BI)) )
	if ((status & (UART_LSR_DR | UART_LSR_BI)) && ((channel_status == 0x1) || (channel_status == 0x2)))
	{
		receive_chars(up, &status);
	}


	check_modem_status(up);

	if (status & UART_LSR_THRE)
	{
		transmit_chars(up);
	}

	spin_unlock_irqrestore(&up->port.lock, flags);
	DEBUG_FUNC_E(__FUNCTION__);
}

/*
 * This is the serial driver's interrupt routine.
 *
 * Arjan thinks the old way was overly complex, so it got simplified.
 * Alan disagrees, saying that need the complexity to handle the weird
 * nature of ISA shared interrupts. (This is a special exception.)
 *
 * In order to handle ISA shared interrupts properly, we need to check
 * that all ports have been serviced, and therefore the ISA interrupt
 * line has been de-asserted.
 *
 * This means we need to loop through all ports. checking that they
 * don't have an interrupt pending.
 */
static irqreturn_t me9000_interrupt(int irq, void *dev_id)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct irq_info *i = dev_id;
	struct list_head *l, *end = NULL;
	int pass_counter = 0, handled = 0;

	DEBUG_INTR("me9000_interrupt(%d) 0x%x -- 0x%x...\n", irq, i->irq, &i->lock);

	spin_lock_bh(&i->lock);
	DEBUG_INTR("Atomic Interrupt start\n");
	DEBUG_INTR("dev_id 0x%x -- 0x%x -- 0x%x\n", i->irq, i->head, &i->lock);

	l = i->head;
	do {
		struct uart_me9000_port *up;
		unsigned int iir, lcr;

		up = list_entry(l, struct uart_me9000_port, list);

		//fab
		long x1=serial_in_l(up, 0x80);
		lcr = serial_in(up, UART_LCR); // store value of LCR
		serial_out(up, UART_LCR, lcr & ~UART_LCR_DLAB); // ensure LCR bit-7=0 before reading UART_IIR
		iir = serial_in(up, UART_IIR);
		serial_out(up, UART_LCR, lcr); // restore LCR

		DEBUG_INTR("me9000_interrupt port=0x%x iir=0x%x - pass_counter=%d\n", up->channelnum, iir, pass_counter);
		if (!(iir & UART_IIR_NO_INT)) {

			DEBUG_INTR("me9000_interrupt handle port\n");

			me9000_handle_port(up);

			handled = 1;

			end = NULL;

		} else if (end == NULL)
			end = l;

		//serial_out(up, UART_LCR, lcr); // restore LCR
		l = l->next;

		if (l == i->head && pass_counter++ > 256) {
			/* If we hit this, we're dead. */
			printk(KERN_ERR "me9000: too much work for irq%d\n", irq);
			break;
		}
	} while (l != end);

	DEBUG_INTR("Atomic Interrupt end\n");
	spin_unlock_bh(&i->lock);

	DEBUG_FUNC_E(__FUNCTION__);

	return IRQ_RETVAL(handled);
}

/*
 * To support ISA shared interrupts, we need to have one interrupt
 * handler that ensures that the IRQ line has been deasserted
 * before returning. Failing to do this will result in the IRQ
 * line being stuck active, and, since ISA irqs are edge triggered,
 * no more IRQs will be seen.
 */
static void serial_do_unlink(struct irq_info *i, struct uart_me9000_port *up)
{
	DEBUG_FUNC_S(__FUNCTION__);
	spin_lock_irq(&i->lock);

	if (!list_empty(i->head)) {
		if (i->head == &up->list)
			i->head = i->head->next;
		list_del(&up->list);
	} else {
		BUG_ON(i->head != &up->list);
		i->head = NULL;
	}

	//fab
	//spin_unlock_irq(&i->lock);

	/* List empty so throw away the hash node */
	if (i->head == NULL) {
		hlist_del(&i->node);
		kfree(i);
	}
	spin_unlock_irq(&i->lock);

	DEBUG_FUNC_E(__FUNCTION__);
}

static int serial_link_irq_chain(struct uart_me9000_port *up)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct hlist_head *h;
	struct hlist_node *n;
	struct irq_info *i;
	int ret, irq_flags = up->port.flags & UPF_SHARE_IRQ ? IRQF_SHARED : 0;

	mutex_lock(&hash_mutex);

	h = &irq_lists[up->port.irq % NR_IRQ_HASH];

	hlist_for_each(n, h) {
		i = hlist_entry(n, struct irq_info, node);
		if (i->irq == up->port.irq)
			break;
	}

	if (n == NULL) {
		i = kzalloc(sizeof(struct irq_info), GFP_KERNEL);
		if (i == NULL) {
			mutex_unlock(&hash_mutex);
			return -ENOMEM;
		}
		spin_lock_init(&i->lock);
		i->irq = up->port.irq;
		hlist_add_head(&i->node, h);
	}
	mutex_unlock(&hash_mutex);

	spin_lock_irq(&i->lock);

	if (i->head) {
		list_add_tail(&up->list, i->head);
		spin_unlock_irq(&i->lock);

		ret = 0;
	} else {
		INIT_LIST_HEAD(&up->list);
		i->head = &up->list;
		spin_unlock_irq(&i->lock);
		irq_flags |= up->port.irqflags;
		//register interrupt handler
		ret = request_irq(up->port.irq, me9000_interrupt, irq_flags, "me9000", i);
		if (ret < 0)
			serial_do_unlink(i, up);
	}

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

static void serial_unlink_irq_chain(struct uart_me9000_port *up)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct irq_info *i;
	struct hlist_node *n;
	struct hlist_head *h;

	mutex_lock(&hash_mutex);

	h = &irq_lists[up->port.irq % NR_IRQ_HASH];

	hlist_for_each(n, h) {
		i = hlist_entry(n, struct irq_info, node);
		if (i->irq == up->port.irq)
			break;
	}

	BUG_ON(n == NULL);
	BUG_ON(i->head == NULL);

	if (list_empty(i->head))
		free_irq(up->port.irq, i);

	serial_do_unlink(i, up);
	mutex_unlock(&hash_mutex);
	DEBUG_FUNC_E(__FUNCTION__);
}

static inline int poll_timeout(int timeout)
{
	DEBUG_FUNC_S(__FUNCTION__);
	DEBUG_FUNC_E(__FUNCTION__);
	return timeout > 6 ? (timeout / 2 - 2) : 1;
}

/*
 * This function is used to handle ports that do not have an
 * interrupt. This doesn't work very well for 16450's, but gives
 * barely passable results for a 16550A. (Although at the expense
 * of much CPU overhead).
 */
 //fab 1903
 //struct timer_list
//static void me9000_timeout(unsigned long data)
// kernel > 4,14,0 (e.g. Ubuntu 16.04.6 LTS) 

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static void me9000_timeout(unsigned long data)
{
	struct uart_me9000_port *up = (struct uart_me9000_port *)data;
#else
static void me9000_timeout(struct timer_list *t)
{
	//pattern: myDevice *dev = from_timer(dev, t, getIntrTimer);
	struct uart_me9000_port *up = from_timer(up, t, timer);
#endif

	DEBUG_FUNC_S(__FUNCTION__);
	unsigned int iir;

	iir = serial_in(up, UART_IIR);
	if (!(iir & UART_IIR_NO_INT))
		me9000_handle_port(up);
	mod_timer(&up->timer, jiffies + poll_timeout(up->port.timeout));
	DEBUG_FUNC_E(__FUNCTION__);
}

static unsigned int me9000_tx_empty(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned long flags;
	unsigned int lsr;

	spin_lock_irqsave(&up->port.lock, flags);
	lsr = serial_in(up, UART_LSR);
	up->lsr_saved_flags |= lsr & LSR_SAVE_FLAGS;
	spin_unlock_irqrestore(&up->port.lock, flags);

	DEBUG_FUNC_E(__FUNCTION__);
	return (lsr & BOTH_EMPTY) == BOTH_EMPTY ? TIOCSER_TEMT : 0;
}

static unsigned int me9000_get_mctrl(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned int status;
	unsigned int ret;

	status = check_modem_status(up);

	ret = 0;
	if (status & UART_MSR_DCD)
		ret |= TIOCM_CAR;
	if (status & UART_MSR_RI)
		ret |= TIOCM_RNG;
	if (status & UART_MSR_DSR)
		ret |= TIOCM_DSR;
	if (status & UART_MSR_CTS)
		ret |= TIOCM_CTS;

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

static void me9000_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned char mcr = 0, efr;

	DEBUG_INTR("me9000_set_mctrl mctrl: 0x%x\n", mctrl);
	if (mctrl & TIOCM_RTS)
		mcr |= UART_MCR_RTS;
	if (mctrl & TIOCM_DTR)
		mcr |= UART_MCR_DTR;
	if (mctrl & TIOCM_OUT1)
		mcr |= UART_MCR_OUT1;
	if (mctrl & TIOCM_OUT2)
		mcr |= UART_MCR_OUT2;
	if (mctrl & TIOCM_LOOP)
		mcr |= UART_MCR_LOOP;

	mcr = (mcr & up->mcr_mask) | up->mcr_force | up->mcr;

	efr = serial_in(up, UART_XR_EFR);
	serial_out(up, UART_XR_EFR, efr & ~UART_EFR_ECB);// clear access to shaded registers so that write to MCR does not change from using DTR to RTS for RS-485 control -> TODO right?
	
	serial_out(up, UART_MCR, mcr);

	serial_out(up, UART_XR_EFR, efr);
	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_break_ctl(struct uart_port *port, int break_state)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned long flags;

	spin_lock_irqsave(&up->port.lock, flags);
	if (break_state == -1)
		up->lcr |= UART_LCR_SBC;
	else
		up->lcr &= ~UART_LCR_SBC;
	serial_out(up, UART_LCR, up->lcr);
	spin_unlock_irqrestore(&up->port.lock, flags);
	DEBUG_FUNC_E(__FUNCTION__);
}

static int me9000_startup(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned long flags;
	unsigned int fctr_reg = 0;
	int retval;

	up->capabilities = uart_config[up->port.type].flags;

	serial_out(up, UART_XR_EFR, UART_EFR_ECB);
	serial_out(up, UART_IER, 0);

	//fab
	DEBUG_INTR(KERN_INFO "Revison/Idend  0x%x -- 0x%x", serial_in(up, XR17V158_DREV), serial_in(up, XR17V158_DVID));
	/* Set the RX/TX trigger levels */
	/* These are some default values, the OEMs can change these values according to their best case scenarios */
	serial_out(up, XR17V158_EXTENDED_RXTRG, 32); //32
	serial_out(up, XR17V158_EXTENDED_TXTRG, 32); //32

	/* Hysteresis level of 8 */
	fctr_reg = serial_in(up, UART_EXAR_FCTR);
	DEBUG_INTR(KERN_INFO "me9000_startup: FCTR=0x%x", fctr_reg);
	//fab
	serial_out(up, UART_EXAR_FCTR, fctr_reg | UART_FCTR_EXAR_TRGD | XR17V158_FCTR_RTS_DELAY);
	//serial_out(up, UART_EXAR_FCTR, 0);

	serial_out(up, UART_LCR, 0);

	/* Wake up and initialize UART */
	serial_out(up, UART_XR_EFR, UART_EFR_ECB); /*Enable Shaded bits access*/
	serial_out(up, UART_MSR, 0);
	serial_out(up, UART_IER, 0);
	serial_out(up, UART_LCR, 0);

	/*
	 * Clear the FIFO buffers and disable them.
	 * (they will be reeanbled in set_termios())
	 */
	serial_out(up, UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
	serial_out(up, UART_FCR, 0);

	/*
	 * Clear the interrupt registers.
	 */
	DEBUG_INTR(KERN_INFO "Clear interrupt registers");
	(void) serial_in(up, UART_LSR);
	(void) serial_in(up, UART_RX);
	(void) serial_in(up, UART_IIR);
	(void) serial_in(up, UART_MSR);

	if (port->irq) {
		retval = serial_link_irq_chain(up);
		if (retval)
			return retval;
	}

	/*
	 * Now, initialize the UART
	 */
	DEBUG_INTR(KERN_INFO "initialize the UART");
	serial_out(up, UART_LCR, UART_LCR_WLEN8);

	spin_lock_irqsave(&up->port.lock, flags);

	/*
	* Most PC uarts need OUT2 raised to enable interrupts.
	*/
	if (is_real_interrupt(up->port.irq))
		up->port.mctrl |= TIOCM_OUT2;
	//to enable intenal loop, uncomment the line below
	//up->port.mctrl |= TIOCM_LOOP;

	me9000_set_mctrl(&up->port, up->port.mctrl);
	spin_unlock_irqrestore(&up->port.lock, flags);

	/*
	 * Finally, enable interrupts. Note: Modem status interrupts
	 * are set via set_termios(), which will be occurring imminently
	 * anyway, so we don't enable them here.
	 */
	up->ier = UART_IER_RLSI | UART_IER_RDI;
	serial_out(up, UART_IER, up->ier);

	/*
	 * And clear the interrupt registers again for luck.
	 */
	DEBUG_INTR(KERN_INFO "Clear interrupt registers again");
	(void) serial_in(up, UART_LSR);
	(void) serial_in(up, UART_RX);
	(void) serial_in(up, UART_IIR);
	(void) serial_in(up, UART_MSR);

	DEBUG_FUNC_E(__FUNCTION__);
	return 0;
}

static void me9000_shutdown(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned long flags;


	/*
	 * Disable interrupts from this port
	 */
	up->ier = 0;
	serial_out(up, UART_IER, 0);

	spin_lock_irqsave(&up->port.lock, flags);

	up->port.mctrl &= ~TIOCM_OUT2;

	me9000_set_mctrl(&up->port, up->port.mctrl);
	spin_unlock_irqrestore(&up->port.lock, flags);

	/*
	 * Disable break condition and FIFOs
	 */
	serial_out(up, UART_LCR, serial_in(up, UART_LCR) & ~UART_LCR_SBC);
	serial_out(up, UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
	serial_out(up, UART_FCR, 0);

	/*
	 * Read data port to reset things, and then unlink from
	 * the IRQ chain.
	 */
	(void) serial_in(up, UART_RX);
#if 1
	if (port->irq)
		serial_unlink_irq_chain(up);
#endif

	DEBUG_FUNC_E(__FUNCTION__);
}

static unsigned int uart_get_divisor_exar(struct uart_port *port, unsigned int baud)
{
	DEBUG_FUNC_S(__FUNCTION__);
	unsigned int quot;

	/*
	 * Old custom speed handling.
	 */
	if (baud == 38400 && (port->flags & UPF_SPD_MASK) == UPF_SPD_CUST)
		quot = port->custom_divisor;
	else
		quot = DIV_ROUND_CLOSEST(port->uartclk, quot_coeff * baud);

	DEBUG_INTR(KERN_INFO "uart_get_divisor_exar: UartClk=%d QuotCoeff=0x%x  baud %d  quot 0x%x\n",port->uartclk,quot_coeff, baud, quot);

	DEBUG_FUNC_E(__FUNCTION__);
	return quot;
}

static unsigned int me9000_get_divisor(struct uart_port *port, unsigned int baud)
{
	DEBUG_FUNC_S(__FUNCTION__);
	unsigned int quot;

	quot = uart_get_divisor_exar(port, baud);

	DEBUG_FUNC_E(__FUNCTION__);
	return quot;
}

static void me9000_set_special_baudrate(struct uart_port *port,unsigned int special_baudrate)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	signed int baud, quot, quot_fraction, port_index;
	unsigned char val_4xmode;
	unsigned char val_8xmode;
	unsigned char lcr_bak;

	printk(KERN_INFO "Enter in me9000_set non-standard baudrate:%d\n", special_baudrate);
	port_index = up->port.line;
	baud = special_baudrate; /*uart_get_baud_rate(port, termios, old, 0, port->uartclk/4)*/
	lcr_bak = serial_in(up, UART_LCR);
	val_4xmode = serial_in(up, XR17V158_4XMODE);
	val_8xmode = serial_in(up, XR17V158_8XMODE);

	if ((port_index > 15) || (port_index < 0))
	{
		return;
	}
	if (port_index >= 8) port_index = port_index - 8;

	if (baud > (port->uartclk / 8)) {
		baud = port->uartclk / 8;
		printk(KERN_INFO "Set baud to max: %d\n", baud);
	}

	if (baud < (port->uartclk / 16)) //781250  (baud < 12500000 / 16)
	{//using the 16x mode
		val_8xmode &= ~(1 << port_index);
		quot_coeff = 16;
		printk(KERN_INFO "Using the 16x Mode\n");
	}
	else  //if ((baud >= 9600) && (baud < 12500000 / 4))  //781250 - 3125000 ((baud >= 12500000 / 16) && (baud < 12500000 / 4))
	{//using the 8x mode
		val_8xmode |= (1 << port_index);
		quot_coeff = 8;
		printk(KERN_INFO "Using the 8x Mode\n");
	}

	DEBUG_INTR(KERN_INFO "write val_8xmode:%d \n", val_8xmode);

	serial_out(up, XR17V158_8XMODE, val_8xmode);
	//serial_out(up, XR17V158_4XMODE, val_4xmode);
	//DEBUG_INTR(KERN_INFO "XR17V158_4XMODE:%d \n", serial_in(up, XR17V158_4XMODE));
	DEBUG_INTR(KERN_INFO "XR17V158_8XMODE:%d \n", serial_in(up, XR17V158_8XMODE));

	quot = me9000_get_divisor(port, baud);
	DEBUG_INTR(KERN_INFO "me9000 uartclk:%d Quot=0x%x\n", port->uartclk,quot);
    //#ifdef DIVISOR_CHANGED
	if ((port->uartclk / baud) > (quot_coeff * quot))
	{
		if (quot_coeff == 16) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot));
		else if (quot_coeff == 8) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 2;
		else if (quot_coeff == 4) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 4;
		DEBUG_INTR(KERN_INFO "quot_fraction (1)=0x%x\n", quot_fraction);
	}
	else if (quot > 1)
	{
		quot--;
		if (quot_coeff == 16) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot));
		else if (quot_coeff == 8) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 2;
		else if (quot_coeff == 4) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 4;
		//fab
		quot++;
		DEBUG_INTR(KERN_INFO "quot_fraction (2)=0x%x\n", quot_fraction);

	}
	else
	{
		quot_fraction = 0;
	}

	if (quot_fraction >= 0x10) quot_fraction = 0x0F;

	serial_out(up, UART_LCR, lcr_bak | UART_LCR_DLAB); /* set DLAB */
	serial_out(up, UART_DLL, quot & 0xFF); /* LS of divisor */
	serial_out(up, UART_DLM, quot >> 8); /* MS of divisor */
	serial_out(up, UART_LCR, lcr_bak); /* reset DLAB */

	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_set_termios(struct uart_port *port, struct ktermios *termios, struct ktermios *old)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	unsigned char cval;
	unsigned long flags;
	signed int baud, quot, quot_fraction;
	unsigned int val_4xmode;
	unsigned int val_8xmode;
	unsigned int reg_read;

	int port_index = up->port.line;
	switch (termios->c_cflag & CSIZE)
	{
		case CS5:
			cval = 0x00;
			break;
		case CS6:
			cval = 0x01;
			break;
		case CS7:
			cval = 0x02;
			break;
		default:
		case CS8:
			cval = 0x03;
			break;
	}

	if (termios->c_cflag & CSTOPB)
		cval |= 0x04;
	if (termios->c_cflag & PARENB)
		cval |= UART_LCR_PARITY;
	if (!(termios->c_cflag & PARODD))
		cval |= UART_LCR_EPAR;
#ifdef CMSPAR
	if (termios->c_cflag & CMSPAR)
		cval |= UART_LCR_SPAR;
#endif

	/*
	 * Ask the core to calculate the divisor for us.
	 */
	//baud = uart_get_baud_rate(port, termios, old, 0, port->uartclk / 4);
	baud = uart_get_baud_rate(port, termios, old, 0, port->uartclk);

	DEBUG_INTR(KERN_INFO "\nme9000_set_termios: Port Index:%d c_ispeed:%d c_ospeed:%d baud=%d\n", port_index, termios->c_ispeed, termios->c_ospeed, baud);

	val_4xmode = 0;
	//val_4xmode = serial_in(up, XR17V158_4XMODE);
	val_8xmode = serial_in(up, XR17V158_8XMODE);


	if ((port_index > 15) || (port_index < 0))
	{
		return;
	}
	if (port_index >= 8) port_index = port_index - 8;

	if (baud > (port->uartclk / 8)) {
		baud = port->uartclk / 8;
		printk(KERN_INFO "Set baud to max: %d\n", baud);
	}

	if (baud < (port->uartclk / 16)) //781250  (baud < 12500000 / 16)
	{//using the 16x mode
		val_8xmode &= ~(1 << port_index);
		quot_coeff = 16;
		printk(KERN_INFO "Using the 16x Mode\n");
	}
	else  //if ((baud >= 9600) && (baud < 12500000 / 4))  //781250 - 3125000 ((baud >= 12500000 / 16) && (baud < 12500000 / 4))
	{//using the 8x mode
		val_8xmode |= (1 << port_index);
		quot_coeff = 8;
		printk(KERN_INFO "Using the 8x Mode\n");
	}


	DEBUG_INTR(KERN_INFO "write val_8xmode:%d \n", val_8xmode);

	serial_out(up, XR17V158_8XMODE, val_8xmode);
	//serial_out(up, XR17V158_4XMODE, val_4xmode);
	//DEBUG_INTR(KERN_INFO "XR17V158_4XMODE:%d \n", serial_in(up, XR17V158_4XMODE));
	DEBUG_INTR(KERN_INFO "XR17V158_8XMODE:%d \n", serial_in(up, XR17V158_8XMODE));

	quot = me9000_get_divisor(port, baud);
	DEBUG_INTR(KERN_INFO "me9000 uartclk:%d Quot=0x%x\n", port->uartclk,quot);
//#ifdef DIVISOR_CHANGED
	if ((port->uartclk / baud) > (quot_coeff * quot))
	{
		if (quot_coeff == 16) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot));
		else if (quot_coeff == 8) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 2;
		else if (quot_coeff == 4) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 4;
		DEBUG_INTR(KERN_INFO "quot_fraction (1)=0x%x\n", quot_fraction);
	}
	else if (quot > 1)
	{
		quot--;
		if (quot_coeff == 16) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot));
		else if (quot_coeff == 8) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 2;
		else if (quot_coeff == 4) quot_fraction = ((port->uartclk / baud) - (quot_coeff * quot)) * 4;
		//fab
		quot++;
		DEBUG_INTR(KERN_INFO "quot_fraction (2)=0x%x\n", quot_fraction);

	}
	else
	{
		quot_fraction = 0;
	}

	if (quot_fraction >= 0x10) quot_fraction = 0x0F;
//#endif
	/*
	 * Ok, we're now changing the port state. Do it with interrupts disabled.
	 */
	spin_lock_irqsave(&up->port.lock, flags);

	/*
	 * Update the per-port timeout.
	 */
	//fab see trace IER/DLM ? --> 0xd
	uart_update_timeout(port, termios->c_cflag, baud);

	up->port.read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_PE | UART_LSR_DR;
	if (termios->c_iflag & INPCK)
		up->port.read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	if (termios->c_iflag & (BRKINT | PARMRK))
		up->port.read_status_mask |= UART_LSR_BI;

	/*
	 * Characteres to ignore
	 */
	up->port.ignore_status_mask = 0;
	if (termios->c_iflag & IGNPAR)
		up->port.ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
	if (termios->c_iflag & IGNBRK) {
		up->port.ignore_status_mask |= UART_LSR_BI;
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns too (for real raw support).
		 */
		if (termios->c_iflag & IGNPAR)
			up->port.ignore_status_mask |= UART_LSR_OE;
	}

	/*
	 * ignore all characters if CREAD is not set
	 */
	if ((termios->c_cflag & CREAD) == 0)
		up->port.ignore_status_mask |= UART_LSR_DR;

	/*
	 * CTS flow control flag and modem status interrupts
	 */
	up->ier &= ~UART_IER_MSI;
	if (UART_ENABLE_MS(&up->port, termios->c_cflag))
		up->ier |= UART_IER_MSI;

	//fab see trace IER/DLM ? --> 0xd
	//serial_out(up, UART_IER, up->ier);

	reg_read = serial_in(up, UART_XR_EFR);

	if (termios->c_cflag & CRTSCTS)
	{
		serial_out(up, UART_XR_EFR, reg_read | UART_EFR_CTS | UART_EFR_RTS);
		printk(KERN_INFO "Hardware Flow Control Enabled\n");
	}
	 else
	 {
		serial_out(up, UART_XR_EFR, reg_read & ~(UART_EFR_CTS | UART_EFR_RTS));
		printk(KERN_INFO "Hardware Flow Control Disabled\n");
	 }

	/*
	* Auto XON/XOFF software flow control flags
	*/
	reg_read = serial_in(up, UART_XR_EFR);
	serial_out(up, XR17V158_UART_XON1, 0x11); //Initializing XON1
	serial_out(up, XR17V158_UART_XOFF1, 0x13); //Initializing XOFF1
	//serial_out(up, XR17V158_UART_XON2, 0x11); //Initializing XON2
	//serial_out(up, XR17V158_UART_XOFF2, 0x13); //Initializing XOFF2

	if (((termios->c_iflag) & IXOFF) && ((termios->c_iflag) & IXON))
	{
		serial_out(up, UART_XR_EFR, (reg_read) | XR17V158_EFR_TRANSMIT_XON1_XOFF1 | XR17V158_EFR_COMPARE_XON1_XOFF1);
		printk(KERN_INFO "Software Flow Control Enabled\n");
	}
	else
	{
		serial_out(up, UART_XR_EFR, (reg_read) & 0xF0);
		printk(KERN_INFO "No Software Flow Control\n");
	}
	
	if ((termios->c_iflag) & IXANY)
	{
		serial_out(up, UART_XR_EFR, ((termios->c_iflag) & IXOFF) || ((termios->c_iflag) & IXON) ? ((reg_read) | UART_EFR_ECB | XR17V158_EFR_TRANSMIT_XON1_XOFF1 | XR17V158_EFR_COMPARE_XON1_XOFF1) : ((reg_read) | UART_EFR_ECB));
		reg_read = serial_in(up, UART_MCR);
		serial_out(up, UART_MCR, (reg_read) | UART_MCR_XONANY);
		serial_out(up, UART_XR_EFR, (reg_read) & ~UART_EFR_ECB);
		printk(KERN_INFO "AUTO XANY Enabled\n");
	}
	else
	{
		serial_out(up, UART_XR_EFR, (reg_read) | UART_EFR_ECB);
		reg_read = serial_in(up, UART_MCR);
		serial_out(up, UART_MCR, (reg_read) & ~UART_MCR_XONANY);
		reg_read = serial_in(up,UART_XR_EFR);
		serial_out(up, UART_XR_EFR, (reg_read) & ~UART_EFR_ECB);
		printk(KERN_INFO "AUTO XANY NOT Enabled\n");
	}

//---------------------------------------------------------------------------//

	serial_out(up, UART_LCR, cval | UART_LCR_DLAB);/* set DLAB */

	DEBUG_INTR(KERN_INFO "quot /old DLM/DLL: 0x%x/0x%x/0x%x\n", quot, serial_in(up, UART_DLM), serial_in(up, UART_DLL));
	serial_out(up, UART_DLL, quot & 0xFF);		/* LS of divisor */
	serial_out(up, UART_DLM, quot >> 8);		/* MS of divisor */
	DEBUG_INTR(KERN_INFO "quot /new DLM/DLL: 0x%x/0x%x/0x%x\n", quot, serial_in(up, UART_DLM), serial_in(up, UART_DLL));
	serial_out(up, UART_LCR, cval);				/* reset DLAB and set cval*/
	up->lcr = cval;								/* Save LCR */

#if FIFO_ON
	//serial_out(up, 0x9, 0x10);/* set fcr */
	serial_out(up, UART_FCR, UART_FCR_ENABLE_FIFO | XR17V158_FCT_RX_FIFO_TRIGGER);/* set fcr */
	//serial_out(up, UART_FCR, 0xc0);
#else
	serial_out(up, UART_FCR, 0);
#endif

	me9000_set_mctrl(&up->port, up->port.mctrl);
#if ENABLE_INTERNAL_LOOPBACK
	reg_read = serial_in(up, UART_MCR);
	serial_out(up, UART_MCR, (reg_read) | UART_MCR_LOOP);
	printk(KERN_INFO "Enabling Internal Loopback\n");
#endif
	spin_unlock_irqrestore(&up->port.lock, flags);

	DEBUG_FUNC_E(__FUNCTION__);
}

static int me9000_get_mpio_config(struct uart_port *port, int arg)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	int ret = 0;
	unsigned int value, inversion, direction = 0;

	//serial_in(up, XR17V158_MPIOINT_7_0);
	value = serial_in(up, XR17V158_MPIOLVL_7_0); //Read IO value
	//serial_in(up, XR17V158_MPIO3T_7_0);
	inversion = serial_in(up, XR17V158_MPIOINV_7_0); //Read Input inversion
	direction = serial_in(up, XR17V158_MPIOSEL_7_0); //Read IO direction

	if (arg & MPIOVAL) {
		if(value & MPIO0) {
			ret |= MPIO0;
		}
		if(value & MPIO1) {
			ret |= MPIO1;
		}
		if(value & MPIO2) {
			ret |= MPIO2;
		}
		if(value & MPIO3) {
			ret |= MPIO3;
		}
		if(value & MPIO4) {
			ret |= MPIO4;
		}
		if(value & MPIO5) {
			ret |= MPIO5;
		}
		if(value & MPIO6) {
			ret |= MPIO6;
		}
		if(value & MPIO7) {
			ret |= MPIO7;
		}
	} else if (arg & MPIOINV) {
		if(inversion & MPIO0) {
			ret |= MPIO0;
		}
		if(inversion & MPIO1) {
			ret |= MPIO1;
		}
		if(inversion & MPIO2) {
			ret |= MPIO2;
		}
		if(inversion & MPIO3) {
			ret |= MPIO3;
		}
		if(inversion & MPIO4) {
			ret |= MPIO4;
		}
		if(inversion & MPIO5) {
			ret |= MPIO5;
		}
		if(inversion & MPIO6) {
			ret |= MPIO6;
		}
		if(inversion & MPIO7) {
			ret |= MPIO7;
		}
	} else if (arg & MPIODIR) {
		if(direction & MPIO0) {
			ret |= MPIO0;
		}
		if(direction & MPIO1) {
			ret |= MPIO1;
		}
		if(direction & MPIO2) {
			ret |= MPIO2;
		}
		if(direction & MPIO3) {
			ret |= MPIO3;
		}
		if(direction & MPIO4) {
			ret |= MPIO4;
		}
		if(direction & MPIO5) {
			ret |= MPIO5;
		}
		if(direction & MPIO6) {
			ret |= MPIO6;
		}
		if(direction & MPIO7) {
			ret |= MPIO7;
		}
	} else {
		ret = -EINVAL;
	}

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

static int me9000_set_mpio_config(struct uart_port *port, int arg)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	int ret = 0;
	unsigned int value, inversion, direction = 0;
	unsigned long flags;

	//serial_in(up, XR17V158_MPIOINT_7_0);
	value = serial_in(up, XR17V158_MPIOLVL_7_0); //Read IO value
	//serial_in(up, XR17V158_MPIO3T_7_0);
	inversion = serial_in(up, XR17V158_MPIOINV_7_0); //Read Input inversion
	direction = serial_in(up, XR17V158_MPIOSEL_7_0); //Read IO direction

	if (arg & MPIOVAL) {
		if(arg & MPIO0) {
			value |= MPIO0;
		} else {
			value &= ~MPIO0;
		}
		if(arg & MPIO1) {
			value |= MPIO1;
		} else {
			value &= ~MPIO1;
		}
		if(arg & MPIO2) {
			value |= MPIO2;
		} else {
			value &= ~MPIO2;
		}
		if(arg & MPIO3) {
			value |= MPIO3;
		} else {
			value &= ~MPIO3;
		}
		if(arg & MPIO4) {
			value |= MPIO4;
		} else {
			value &= ~MPIO4;
		}
		if(arg & MPIO5) {
			value |= MPIO5;
		} else {
			value &= ~MPIO5;
		}
		if(arg & MPIO6) {
			value |= MPIO6;
		} else {
			value &= ~MPIO6;
		}
		if(arg & MPIO7) {
			value |= MPIO7;
		} else {
			value &= ~MPIO7;
		}
	} else if (arg & MPIOINV) {
		if(arg & MPIO0) {
			inversion |= MPIO0;
		} else {
			inversion &= ~MPIO0;
		}
		if(arg & MPIO1) {
			inversion |= MPIO1;
		} else {
			inversion &= ~MPIO1;
		}
		if(arg & MPIO2) {
			inversion |= MPIO2;
		} else {
			inversion &= ~MPIO2;
		}
		if(arg & MPIO3) {
			inversion |= MPIO3;
		} else {
			inversion &= ~MPIO3;
		}
		if(arg & MPIO4) {
			inversion |= MPIO4;
		} else {
			inversion &= ~MPIO4;
		}
		if(arg & MPIO5) {
			inversion |= MPIO5;
		} else {
			inversion &= ~MPIO5;
		}
		if(arg & MPIO6) {
			inversion |= MPIO6;
		} else {
			inversion &= ~MPIO6;
		}
		if(arg & MPIO7) {
			inversion |= MPIO7;
		} else {
			inversion &= ~MPIO7;
		}
	} else if (arg & MPIODIR) {
		if(arg & MPIO0) {
			direction |= MPIO0;
		} else {
			direction &= ~MPIO0;
		}
		if(arg & MPIO1) {
			direction |= MPIO1;
		} else {
			direction &= ~MPIO1;
		}
		if(arg & MPIO2) {
			direction |= MPIO2;
		} else {
			direction &= ~MPIO2;
		}
		if(arg & MPIO3) {
			direction |= MPIO3;
		} else {
			direction &= ~MPIO3;
		}
		if(arg & MPIO4) {
			direction |= MPIO4;
		} else {
			direction &= ~MPIO4;
		}
		if(arg & MPIO5) {
			direction |= MPIO5;
		} else {
			direction &= ~MPIO5;
		}
		if(arg & MPIO6) {
			direction |= MPIO6;
		} else {
			direction &= ~MPIO6;
		}
		if(arg & MPIO7) {
			direction |= MPIO7;
		} else {
			direction &= ~MPIO7;
		}
	} else {
		ret = -EINVAL;
	}

	spin_lock_irqsave(&up->port.lock, flags);
	//serial_out(up, XR17V158_MPIOINT_7_0, 0x00);
	serial_out(up, XR17V158_MPIOLVL_7_0, value); //Set IO value
	//serial_out(up, XR17V158_MPIO3T_7_0, 0x00);
	serial_out(up, XR17V158_MPIOINV_7_0, inversion); //Set Input inversion
	serial_out(up, XR17V158_MPIOSEL_7_0, direction); //Set IO direction
	spin_unlock_irqrestore(&up->port.lock, flags);

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

/*
 * This function is used to handle Exar Device specific ioctl calls
 * The user level application should have defined the above ioctl
 * commands with the above values to access these ioctls and the
 * input parameters for these ioctls should be struct meioctl_rw_reg
 * The Ioctl functioning is pretty much self explanatory here in the code,
 * and the register values should be between 0 to XR17V158_EXTENDED_RXTRG
 */
static int me9000_ioctl(struct uart_port *port, unsigned int cmd, unsigned long arg)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;
	int ret = -ENOIOCTLCMD;
	struct meioctl_rw_reg ioctlrwarg;
	unsigned int any_baudrate = 0;
	int mpio = 0;
	DEBUG_INTR(KERN_INFO "me9000_ioctl cmd: 0x%x\n", cmd);
	switch (cmd)
	{
		case TIOCGMPIO:
			get_from_user(mpio, (int *)arg);
			ret = me9000_get_mpio_config(port, mpio);
			printk(KERN_INFO "me9000_ioctl: TIOCGMPIO arg=0x%x", mpio);
			break;

		case TIOCSMPIO:
			get_from_user(mpio, (int *)arg);
			ret = me9000_set_mpio_config(port, mpio);
			printk(KERN_INFO "me9000_ioctl: TIOCSMPIO arg=0x%x", mpio);
			break;

		case EXAR_READ_REG:
			printk(KERN_INFO "me9000_ioctl read reg[0x%02x]=0x%02x \n", ioctlrwarg.reg, ioctlrwarg.regvalue);
			if (copy_from_user(&ioctlrwarg, (void *)arg, sizeof(ioctlrwarg)))
				return -EFAULT;
			ioctlrwarg.regvalue = serial_in(up, ioctlrwarg.reg);
			printk(KERN_INFO "me9000_ioctl read reg[0x%02x]=0x%02x \n", ioctlrwarg.reg, ioctlrwarg.regvalue);
			if (copy_to_user((void *)arg, &ioctlrwarg, sizeof(ioctlrwarg)))
				return -EFAULT;

			ret = 0;
			break;

		case EXAR_WRITE_REG:
			if (copy_from_user(&ioctlrwarg, (void *)arg, sizeof(ioctlrwarg)))
				return -EFAULT;
			serial_out(up, ioctlrwarg.reg, ioctlrwarg.regvalue);
			DEBUG_INTR(KERN_INFO "me9000_ioctl write reg[0x%02x]=0x%02x \n", ioctlrwarg.reg, ioctlrwarg.regvalue);
			ret = 0;
			break;

		case EXAR_SET_NON_STANDARD_BAUDRATE:
			if (copy_from_user(&any_baudrate, (void *)arg, sizeof(unsigned int)))
			{
				return -EFAULT;
			}
			me9000_set_special_baudrate(port, any_baudrate);
			break;

		case 0x5495:
			DEBUG_INTR(KERN_INFO "me9000_ioctl set rs485 \n");
			struct serial_rs485 rs485;
			if (copy_from_user(&rs485, (void *)arg, sizeof(rs485))) return -EFAULT;
//fab 1903
			//ret = pci_me9000_rs485_config(up, &rs485);
			ret = pci_me9000_rs485_config(port, &rs485);

			break;


	}

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

static void me9000_pm(struct uart_port *port, unsigned int state, unsigned int oldstate)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *up = (struct uart_me9000_port *)port;

	if (state) {
		/* sleep */
		serial_out(up, XR17V158_SLEEP, 0xff);
	} else {
		/* wake */
		serial_out(up, XR17V158_SLEEP, 0x00);
	}

	if (up->pm)
		up->pm(port, state, oldstate);
	DEBUG_FUNC_E(__FUNCTION__);
}

static void me9000_release_port(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	DEBUG_FUNC_E(__FUNCTION__);
}

static int me9000_request_port(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	DEBUG_FUNC_E(__FUNCTION__);
	return 0;
}

static void me9000_config_port(struct uart_port *port, int flags)
{
	DEBUG_FUNC_S(__FUNCTION__);

	struct uart_me9000_port *up = (struct uart_me9000_port *)port;

	if (flags & UART_CONFIG_TYPE)
	{

		up->port.type = XR17V158;
		up->port.fifosize = uart_config[up->port.type].dfl_xmit_fifo_size;
		up->capabilities = uart_config[up->port.type].flags;
		DEBUG_INTR(KERN_INFO "me9000_config_port type 0x%x  fifosize 0x%x capabilities 0x%x\n", up->port.type, up->port.fifosize, up->capabilities);
	}
	DEBUG_FUNC_E(__FUNCTION__);
}

static const char * me9000_type(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	int type = port->type;

	if (type >= ARRAY_SIZE(uart_config))
		type = 0;

	DEBUG_FUNC_E(__FUNCTION__);
	return uart_config[type].name;
}

static struct uart_ops me9000_pops = {
	.tx_empty = me9000_tx_empty,
	.set_mctrl = me9000_set_mctrl,
	.get_mctrl = me9000_get_mctrl,
	.stop_tx = me9000_stop_tx,
	.start_tx = me9000_start_tx,
	.stop_rx = me9000_stop_rx,
	.enable_ms = me9000_enable_ms,
	.break_ctl = me9000_break_ctl,
	.startup = me9000_startup,
	.shutdown = me9000_shutdown,
	.set_termios = me9000_set_termios,
	.pm = me9000_pm,
	.type = me9000_type,
	.release_port = me9000_release_port,
	.request_port = me9000_request_port,
	.config_port = me9000_config_port,
	.ioctl = me9000_ioctl,
};

static struct uart_me9000_port *me9000_find_match_or_unused(struct uart_port *port)
{
	DEBUG_FUNC_S(__FUNCTION__);
	int i;

	/*
	 * First, find a port entry which matches.
	 */
	for (i = 0; i < NR_PORTS; i++)
		if (uart_match_port(&me9000_ports[i].port, port))
		{
			DEBUG_INTR("MatchA %d\n", i);
			DEBUG_FUNC_E(__FUNCTION__);
			return &me9000_ports[i];
		}
	/*
	 * We didn't find a matching entry, so look for the first
	 * free entry. We look for one which hasn't been previously
	 * used (indicated by zero iobase).
	 */
	for (i = 0; i < NR_PORTS; i++)
		if (me9000_ports[i].port.type == PORT_UNKNOWN && me9000_ports[i].port.iobase == 0)
		{
			port->line = i;
			DEBUG_INTR("MatchB %d\n", i);
			DEBUG_FUNC_E(__FUNCTION__);
			return &me9000_ports[i];
		}

	/*
	 * That also failed. Last resort is to find any entry which
	 * doesn't have a real port associated with it.
	 */
	for (i = 0; i < NR_PORTS; i++)
		if (me9000_ports[i].port.type == PORT_UNKNOWN)
		{
			DEBUG_INTR("MatchC %d\n", i);
			DEBUG_FUNC_E(__FUNCTION__);
			return &me9000_ports[i];
		}

	DEBUG_FUNC_E(__FUNCTION__);
	return NULL;
}

/*
 * me9000_register_port - register a serial port
 * @port: serial port template
 *
 * Configure the serial port specified by the request. If the
 * port exists and is in use, it is hung up and unregistered
 * first.
 *
 * The port is then probed and if necessary the IRQ is autodetected
 * If this fails an error is returned.
 *
 * On success the port is ready to use and the line number is returned.
 */
int me9000_register_port(struct uart_port *port, unsigned short deviceid, unsigned char channelnum)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *uart;
	int ret = -ENOSPC;

	if (port->uartclk == 0)
		return -EINVAL;

	mutex_lock(&serial_mutex);
	uart = me9000_find_match_or_unused(port);
	if (uart) {
		uart->port.iobase = port->iobase;
		uart->port.membase = port->membase;
		uart->port.irq = port->irq;
		uart->port.uartclk = port->uartclk;
		uart->port.fifosize = port->fifosize;
		uart->port.regshift = port->regshift;
		uart->port.iotype = port->iotype;
		uart->port.flags = port->flags | UPF_BOOT_AUTOCONF;
		uart->port.mapbase = port->mapbase;
		if (port->dev)
			uart->port.dev = port->dev;

		uart->deviceid = deviceid;
		uart->channelnum = channelnum;
		uart->port.line = port->line;
		//fab
		uart->port.rs485 = port->rs485;
		uart->port.rs485_config = port->rs485_config;


		spin_lock_init(&uart->port.lock);
 //fab 1903 
 // kernel > 4,14,0 (e.g. Ubuntu 16.04.6 LTS) 
 // new API: void timer_setup(struct timer_list *timer, void (*callback)(struct timer_list *), unsigned int flags);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
		init_timer(&uart->timer);
		uart->timer.function = me9000_timeout;
#else		
		timer_setup(&uart->timer, me9000_timeout ,0);
#endif

		/*
		 * ALPHA_KLUDGE_MCR needs to be killed.
		 */
		uart->mcr_mask = ~(0x0); //~ALPHA_KLUDGE_MCR;
		uart->mcr_force = 0; // ALPHA_KLUDGE_MCR;

		uart->port.ops = &me9000_pops;

		ret = uart_add_one_port(&me9000_uart_driver, &uart->port);
#if 0
		if (ret == 0)
		{
			ret = uart->port.line;

			if (is_real_interrupt(uart->port.irq)) {
				serial_link_irq_chain(uart);
			}
		}
#endif
	}
	mutex_unlock(&serial_mutex);

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

/*
 * Probe one serial board. Unfortunately, there is no rhyme nor reason
 * to the arrangement of serial ports on a PCI card.
 */


static int __devinit init_one_me9000(struct pci_dev *dev, const struct pci_device_id *ent)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct serial_private *priv;
	struct pciserial_board *board;
	struct pci_serial_quirk *quirk;
	struct uart_port serial_port;
	int rc, nr_ports, i;

	if (ent->driver_data >= ARRAY_SIZE(mepciserial_boards)) {
		printk(KERN_INFO "init_one_me9000: invalid driver_data: %ld\n", ent->driver_data);
		return -EINVAL;
	}

	board = &mepciserial_boards[ent->driver_data];

	rc = pci_enable_device(dev);
	if (rc)
		return rc;

	nr_ports = board->num_ports;

	/*
	 * Find an init and setup quirks.
	 */
	quirk = find_quirk(dev);

	/*
	 * Run the new-style initialization function.
	 * The initialization function returns:
	 * <0 - error
	 * 0 - use board->num_ports
	 * >0 - number of ports
	 */
	if (quirk->init) {
		rc = quirk->init(dev);
		if (rc < 0)
			goto disable;
		if (rc)
			nr_ports = rc;
	}

	size_t needed_memory = sizeof(struct serial_private) + sizeof(unsigned int) * nr_ports;

	DEBUG_INTR("MemAlloc:  0x%x + 0x%x = 0x%x\n", sizeof(struct serial_private), sizeof(unsigned int) * nr_ports, needed_memory);

	priv = kmalloc(needed_memory, GFP_KERNEL);
	if (!priv) {
		rc = -ENOMEM;
		goto deinit;
	}

	memset(priv, 0, needed_memory);

	priv->dev = dev;
	priv->quirk = quirk;

	memset(&serial_port, 0, sizeof(struct uart_port));
	serial_port.flags = UPF_SKIP_TEST | UPF_BOOT_AUTOCONF | UPF_SHARE_IRQ;

	serial_port.uartclk = board->base_baud * 16;

	serial_port.irq = dev->irq;
	serial_port.dev = &dev->dev;

	for (i = 0; i < nr_ports; i++) {
		if (quirk->setup(priv, board, &serial_port, i))
			break;

		rc = me9000_register_port(&serial_port, dev->device, i);
		if (rc < 0) {
			printk(KERN_WARNING "Couldn't register serial port %s: %d\n", pci_name(dev), i);
			break;
		}

		printk(KERN_WARNING "init_one_me9000 line:%d\n", serial_port.line);
		priv->uart_index[i] = serial_port.line;
		priv->line[i] = rc;
	}

	priv->nr = i;

	if (!IS_ERR(priv)) {
		pci_set_drvdata(dev, priv);
		DEBUG_FUNC_E(__FUNCTION__);
		return 0;
	}

deinit:
	if (quirk->exit)
		quirk->exit(dev);
disable:
	pci_disable_device(dev);

	DEBUG_INTR(KERN_INFO "Error rc=0x%x\n", rc);
	DEBUG_FUNC_E(__FUNCTION__);
	return rc;
}

/*
 * me9000_unregister_port - remove a serial port at runtime
 * @line: serial line number
 *
 * Remove one serial port. This may not be called from interrupt
 * context. We hand the port back to the our control.
 */
void me9000_unregister_port(int line)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct uart_me9000_port *uart = &me9000_ports[line];

	mutex_lock(&serial_mutex);
#if 0
	if (is_real_interrupt(uart->port.irq))
		serial_unlink_irq_chain(uart);
#endif
	uart_remove_one_port(&me9000_uart_driver, &uart->port);
	uart->port.dev = NULL;
	mutex_unlock(&serial_mutex);
	DEBUG_FUNC_E(__FUNCTION__);
}

void me9000_remove_ports(struct serial_private *priv)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct pci_serial_quirk *quirk;
	int i;

	for (i = 0; i < priv->nr; i++)
	{
		printk(KERN_WARNING "me9000_remove_ports dev:%p port_num:%d\n", priv->dev, priv->uart_index[i]);
		//me9000_unregister_port(priv->line[i]);
		me9000_unregister_port(priv->uart_index[i]);

	}

	for (i = 0; i < PCI_NUM_BAR_RESOURCES; i++) {
		if (priv->remapped_bar[i])
			iounmap(priv->remapped_bar[i]);
		priv->remapped_bar[i] = NULL;
	}

	/*
	 * Find the exit quirks.
	 */
	quirk = find_quirk(priv->dev);
	if (quirk->exit)
		quirk->exit(priv->dev);

	kfree(priv);
	DEBUG_FUNC_E(__FUNCTION__);
}

static void __devexit remove_one_me9000(struct pci_dev *dev)
{
	DEBUG_FUNC_S(__FUNCTION__);
	struct serial_private *priv = pci_get_drvdata(dev);

	pci_set_drvdata(dev, NULL);

	me9000_remove_ports(priv);

	pci_disable_device(dev);
	DEBUG_FUNC_E(__FUNCTION__);
}

static struct pci_driver me9000_pci_driver = {
	.name = "me9000",
	.probe = init_one_me9000,
	.remove = __devexit_p(remove_one_me9000),
	.id_table = me9000_pci_tbl,
};

static int __init me9000_init(void)
{
	DEBUG_FUNC_S(__FUNCTION__);
	int ret;

	printk(KERN_INFO "Meilhaus ME-9000 driver - Revision: 1.0.7\n");

	ret = uart_register_driver(&me9000_uart_driver);
	if (ret)
		return ret;

	ret = pci_register_driver(&me9000_pci_driver);

	if (ret < 0)
		uart_unregister_driver(&me9000_uart_driver);

	DEBUG_FUNC_E(__FUNCTION__);
	return ret;
}

static void __exit me9000_exit(void)
{
	DEBUG_FUNC_S(__FUNCTION__);
	pci_unregister_driver(&me9000_pci_driver);
	uart_unregister_driver(&me9000_uart_driver);
	DEBUG_FUNC_E(__FUNCTION__);
}

module_init(me9000_init);
module_exit(me9000_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Meilhaus ME-9000 driver - Revision: 1.0.8");
